from geoformat_lib.conversion.precision_tolerance_conversion import (
    deduce_rounding_value_from_float,
    deduce_precision_from_round
)

from tests.test_all import test_function

deduce_rounding_value_from_float_parameters = {
    0: {
        'float_value': 56.4325,
        'return_value': 4
    },
    1: {
        'float_value': 56.43256436,
        'return_value': 8
    },
    2: {
        'float_value': -56.4325,
        'return_value': 4
    },
    3: {
        'float_value': -56.43256436,
        'return_value': 8
    }
}

deduce_precision_from_round_parameters = {
    0: {
        'rounding_value': 4,
        'return_value': 0.0001,
    },
    1: {
        'rounding_value': 8,
        'return_value': 0.00000001
    },
    2: {
        'rounding_value': 4,
        'return_value': 0.0001,
    },
    3: {
        'rounding_value': 8,
        'return_value': 0.00000001,
    }
}


def test_all():
    # deduce_rounding_value_from_float
    print(test_function(deduce_rounding_value_from_float, deduce_rounding_value_from_float_parameters))

    # deduce_precision_from_round
    print(test_function(deduce_precision_from_round, deduce_precision_from_round_parameters))

if __name__ == '__main__':
    test_all()



from geoformat_lib.geoprocessing.connectors.operations import coordinates_to_bbox
from geoformat_lib.geoprocessing.geoparameters.bbox import bbox_union


def get_geocontainer_extent(geocontainer):
    if 'metadata' in geocontainer.keys():
        if 'extent' in geocontainer['metadata']:
            return geocontainer['metadata']['extent']
    else:
        for i_layer, geolayer in enumerate(geocontainer['layers']):
            geolayer_bbox = None
            if 'geometry_ref' in geolayer['metadata']:
                if 'extent' in geolayer['metadata']['geometry_ref']:
                    geolayer_bbox = geolayer['metadata']['geometry_ref']['extent']
            else:
                # no geometry in geolayer
                return None

            if not geolayer_bbox:
                for i_feat in geolayer['features']:
                    feature_bbox = None
                    if 'geometry' in geolayer['features'][i_feat].keys():
                        if 'bbox' in geolayer['features'][i_feat]['geometry']:
                            feature_bbox = geolayer['features'][i_feat]['geometry']['bbox']

                        else:
                            feature_bbox = coordinates_to_bbox(geolayer['features'][i_feat]['geometry'])
                    else:
                        # no geometry in geolayer
                        return None

                    if i_feat == 0:
                        geolayer_bbox = feature_bbox
                    else:
                        geolayer_bbox = bbox_union(geolayer_bbox, feature_bbox)

            if i_layer == 0:
                geocontainer_extent = geolayer_bbox
            else:
                geocontainer_extent = bbox_union(geocontainer_extent, geolayer_bbox)

    return geocontainer_extent



from geoformat_lib.geoprocessing.measure.area import (
    shoelace_formula
)

from tests.test_all import test_function

shoelace_formula_parameters = {
    0: {
        "ring_coordinates": [[-1, 1], [1, 1], [1, -1], [-1, -1]],  # convex boundary
        "absolute_value": True,
        "return_value": 4.
    },
    1: {
        "ring_coordinates": [[-1, -1], [1, -1], [1, 1], [-1, 1]],  # reverse convex boundary
        "absolute_value": True,
        "return_value": 4.
    },
    2: {
        "ring_coordinates": [[-1, -1], [0, 0], [1, -1], [1, 1], [-1, 1]],  # concave boundary
        "absolute_value": True,
        "return_value": 3.
    },
    3: {
        "ring_coordinates": [[-1, 1], [1, 1], [1, -1], [0, 0], [-1, -1]],  # reverse concave boundary
        "absolute_value": True,
        "return_value": 3.
    },
    4: {
        "ring_coordinates": [[-1, 1], [1, 1], [1, -1], [-1, -1]],  # convex boundary
        "absolute_value": False,
        "return_value": 4.
    },
    5: {
        "ring_coordinates": [[-1, -1], [1, -1], [1, 1], [-1, 1]],  # reverse convex boundary
        "absolute_value": False,
        "return_value": -4.
    },
    6: {
        "ring_coordinates": [[-1, -1], [0, 0], [1, -1], [1, 1], [-1, 1]],  # concave boundary
        "absolute_value": False,
        "return_value": -3.
    },
    7: {
        "ring_coordinates": [[-1, 1], [1, 1], [1, -1], [0, 0], [-1, -1]],  # reverse concave boundary
        "absolute_value": False,
        "return_value": 3.
    },
}


def test_all():
    # shoelace_formula
    print(test_function(shoelace_formula, shoelace_formula_parameters))


if __name__ == '__main__':
    test_all()



from geoformat_lib.obj.geometry import (
    len_coordinates,
    len_coordinates_in_geometry
)

from tests.data.coordinates import (
    point_coordinates,
    linestring_coordinates,
    polygon_coordinates,
    multipoint_coordinates,
    multilinestring_coordinates,
    multipolygon_coordinates
)

from tests.data.geometries import (
    POINT,
    LINESTRING,
    POLYGON,
    MULTIPOINT,
    MULTILINESTRING,
    MULTIPOLYGON,
    GEOMETRYCOLLECTION
)

from tests.test_all import test_function

len_coordinates_parameters = {
    0: {
        "coordinates": point_coordinates,
        "return_value": 1
    },
    1: {
        "coordinates": linestring_coordinates,
        "return_value": 2
    },
    2: {
        "coordinates": polygon_coordinates,
        "return_value": 8
    },
    3: {
        "coordinates": multipoint_coordinates,
        "return_value": 3
    },
    4: {
        "coordinates": multilinestring_coordinates,
        "return_value": 5
    },
    5: {
        "coordinates": multipolygon_coordinates,
        "return_value": 8
    }
}

len_coordinates_in_geometry_parameters = {
    0: {
        "geometry": POINT,
        "return_value": 1
    },
    1: {
        "geometry": LINESTRING,
        "return_value": 2
    },
    2: {
        "geometry": POLYGON,
        "return_value": 8
    },
    3: {
        "geometry": MULTIPOINT,
        "return_value": 3
    },
    4: {
        "geometry": MULTILINESTRING,
        "return_value": 5
    },
    5: {
        "geometry": MULTIPOLYGON,
        "return_value": 8
    },
    6: {
        "geometry": GEOMETRYCOLLECTION,
        "return_value": 27
    },

}


def test_all():

    # len_coordinates
    print(test_function(len_coordinates, len_coordinates_parameters))

    # len_coordinates_in_geometry
    print(test_function(len_coordinates_in_geometry, len_coordinates_in_geometry_parameters))


if __name__ == '__main__':
    test_all()


from pathlib import Path

from geoformat_lib.conf.error_messages import path_not_valid, path_not_valid_file_exists_overwrite_is_false


def add_extension_path(path, add_extension):
    """
    Add extension to path if add extension is specified

    :param path: input path
    :param add_extension:
    :return: output path
    """
    if add_extension and add_extension != path.suffix:
        path = path.with_suffix(path.suffix + add_extension)

    return path


def verify_input_path_is_file(path):
    """
    Transform str path to Path object from pathlib if given path exists and is a file
    :param path: str path or pathlib Path object
    :return: Path object
    """
    p = Path(path)
    if p.is_file():
        return p
    else:
        raise Exception(path_not_valid.format(path=path))


def path_to_file_path(path, geolayer_name, overwrite=True, add_extension=None):
    """
    Return verified file path for input path.
    Make difference between path dir and path file and return always

    :param path: dir path or file path to be checked.
    :param geolayer_name: name of geolayer (only used when path is dir path)
    :param overwrite: True we overwrite file if exists, False we return an error if file exists
    :param add_extension: if input path if file (without extension) or dir if add extension is True we add file
    extension specified ('geojson', 'shp', 'kml' ...)
    :return: output path
    """
    p = Path(path)
    file_path = None
    if p.exists() is True:
        # check if is dir
        if p.is_dir() is True:
            file_path = p.joinpath(geolayer_name)
        # chek if it's file
        elif p.is_file():
            file_path = p
    # file or dir does not exists
    else:
        if p.parent.exists():
            file_path = p
        else:
            raise Exception(path_not_valid.format(path=path))

    # add extension
    file_path = add_extension_path(path=file_path, add_extension=add_extension)

    # check overwrite
    if file_path.exists() and overwrite is False:
        raise Exception(path_not_valid_file_exists_overwrite_is_false.format(path=file_path))

    return file_path



import copy

from geoformat_lib.conf.format_data import value_to_iterable_value
from geoformat_lib.conversion.geometry_conversion import (
    multi_geometry_to_single_geometry,
    geometry_type_to_2d_geometry_type,
    geometry_to_2d_geometry,
    reproject_geometry
)
from geoformat_lib.geoprocessing.geoparameters.bbox import (
    bbox_union
)

from geoformat_lib.conversion.feature_conversion import feature_serialize


def multi_geometry_to_single_geometry_geolayer(geolayer):
    # creation de l'output en copiant les metadata de l'input
    geolayer_out = copy.deepcopy(geolayer)

    del geolayer_out['features']
    if geolayer['metadata']['geometry_ref']['type'].upper() == 'MULTIPOINT':  # 'MultiPoint'
        geolayer_out['metadata']['geometry_ref']['type'] = 'Point'
    elif geolayer['metadata']['geometry_ref']['type'].upper() == 'MULTILINESTRING':  # 'MultiLineString'
        geolayer_out['metadata']['geometry_ref']['type'] = 'LineString'
    elif geolayer['metadata']['geometry_ref']['type'].upper() == 'MULTIPOLYGON':  # 'MultiPolygon'
        geolayer_out['metadata']['geometry_ref']['type'] = 'Polygon'

    # boucle et transformation des géométries multi part en single part
    new_i_feat = 0
    geolayer_out['features'] = {}
    for i_feat in geolayer['features']:

        feature = geolayer['features'][i_feat]
        # if feature is serialized
        if 'feature_serialize' in geolayer['metadata']:
            if geolayer['metadata']['feature_serialize']:
                feature = eval(feature)

        geometry = feature['geometry']
        for new_geometry in multi_geometry_to_single_geometry(geometry):
            new_feature = {'attributes': feature['attributes'],
                           'geometry': new_geometry}

            # if feature is serialized
            if 'feature_serialize' in geolayer['metadata']:
                if geolayer['metadata']['feature_serialize']:
                    new_feature = str(new_feature)

            geolayer_out['features'][new_i_feat] = new_feature
            new_i_feat += 1

    return geolayer_out


def geolayer_to_2d_geolayer(input_geolayer):
    """

    :param input_geolayer:
    :return:
    """
    new_geolayer = {'features': {}, 'metadata': copy.deepcopy(input_geolayer['metadata'])}
    input_geometry_type = input_geolayer['metadata']['geometry_ref']['type']
    if isinstance(input_geometry_type, (list, tuple)):
        new_geometry_type = []
        for geom_type in input_geometry_type:
            new_geometry_type.append(geometry_type_to_2d_geometry_type(geom_type))
    else:
        new_geometry_type = geometry_type_to_2d_geometry_type(input_geometry_type)
    new_geolayer['metadata']['geometry_ref']['type'] = new_geometry_type

    if 'extent' in new_geolayer['metadata']['geometry_ref']:
        bbox_extent = True
    else:
        bbox_extent = False

    for i_feat in input_geolayer['features']:
        input_feature = input_geolayer['features'][i_feat]
        if 'feature_serialize' in input_geolayer['metadata']:
            if input_geolayer['metadata']['feature_serialize'] == True:
                input_feature = eval(input_feature)

        output_feature = copy.deepcopy(input_feature)

        if 'geometry' in input_feature:
            input_geometry = input_feature['geometry']
            new_geometry = geometry_to_2d_geometry(input_geometry, bbox=bbox_extent)
            output_feature['geometry'] = new_geometry

        if 'feature_serialize' in input_geolayer['metadata']:
            if input_geolayer['metadata']['feature_serialize'] == True:
                output_feature = str(output_feature)

        new_geolayer['features'][i_feat] = output_feature

    return new_geolayer


def create_geolayer_from_i_feat_list(geolayer, i_feat_list, serialize=False, reset_i_feat=True):
    """
    Create a new layer with i_feat_list from an input layer
    """
    i_feat_list = value_to_iterable_value(value=i_feat_list, output_iterable_type=list)

    new_layer = {
        'metadata': dict(geolayer['metadata'])
    }

    if serialize:
        geolayer['metadata']['feature_serialize'] = True

    new_layer['features'] = {}
    for new_i_feat, i_feat in enumerate(i_feat_list):
        if i_feat in geolayer['features']:
            if serialize:
                new_feature = feature_serialize(geolayer['features'][i_feat])
            else:
                new_feature = geolayer['features'][i_feat]

        if reset_i_feat:
            new_layer['features'][new_i_feat] = new_feature
        else:
            new_layer['features'][i_feat] = new_feature

    return new_layer


def reproject_geolayer(geolayer, out_crs, in_crs=None, bbox=True):
    """
    Reproject geolayer from crs to an other

    :param geolayer: input geolayer.
    :param out_crs: new crs.
    :param in_crs:  input geolayer crs (if not present in geolayer metadata).
    :param bbox: add bbox to geolayer.
    :return:
    """

    geolayer = copy.deepcopy(geolayer)
    if not in_crs:
        in_crs = geolayer['metadata']['geometry_ref']['crs']

    # change metadata
    geolayer['metadata']['geometry_ref']['crs'] = out_crs

    if bbox:
        geolayer['metadata']['geometry_ref']['extent'] = None

    # reproject geometry
    for i_feat in geolayer['features']:
        feature = geolayer['features'][i_feat]

        # if geometry in feature
        if 'geometry' in feature.keys():
            feature_geometry = feature['geometry']
            new_geometry = reproject_geometry(
                geometry=feature_geometry,
                in_crs=in_crs,
                out_crs=out_crs,
                bbox_extent=bbox
            )

            if geolayer['metadata']['geometry_ref']['extent'] is None:
                geolayer['metadata']['geometry_ref']['extent'] = new_geometry['bbox']
            else:
                bbox_extent = geolayer['metadata']['geometry_ref']['extent']
                geolayer['metadata']['geometry_ref']['extent'] = bbox_union(bbox_extent, new_geometry['bbox'])

            # assign new geometry
            feature['geometry'] = new_geometry

    if bbox:
        geolayer['metadata']['geometry_ref']['extent'] = bbox_extent

    return geolayer



from geoformat_lib.geoprocessing.matrix.adjacency import (
    create_adjacency_matrix,
    get_neighbor_i_feat,
    get_area_intersecting_neighbors_i_feat
)

from tests.test_all import (
    test_function,
    test_dependencies
)

from tests.data.geolayers import geolayer_grid_3_3
from tests.data.matrix import matrix_grid_3_3

create_adjacency_matrix_parameters = {
    0:  {
        'geolayer': geolayer_grid_3_3,
        'mesh_size': None,
        'return_value': matrix_grid_3_3
    }
}

get_neighbor_i_feat_parameters = {
    0: {
        "adjacency_matrix": matrix_grid_3_3,
        "i_feat": 4,
        "return_value": {0, 1, 2, 3, 5, 6, 7, 8}
    },
    1: {
        "adjacency_matrix": matrix_grid_3_3,
        "i_feat": 8,
        "return_value": {4, 5, 7,}
    },
}

get_area_intersecting_neighbors_i_feat_parameters = {
    0: {
        'adjacency_matrix': matrix_grid_3_3,
        'i_feat': 4,
        'neighbor_set': None,
        "return_value": {0, 1, 2, 3, 4, 5, 6, 7, 8}
    },
    1: {
        'adjacency_matrix': matrix_grid_3_3,
        'i_feat': 0,
        'neighbor_set': None,
        "return_value": {0, 1, 2, 3, 4, 5, 6, 7, 8}
    }
}


def test_all():

    if test_dependencies()['ogr']:
        # create_adjacency_matrix
        print(test_function(create_adjacency_matrix, create_adjacency_matrix_parameters))

    # get_neighbor_i_feat
    print(test_function(get_neighbor_i_feat, get_neighbor_i_feat_parameters))

    # get_area_intersecting_neighbors_i_feat
    print(test_function(get_area_intersecting_neighbors_i_feat, get_area_intersecting_neighbors_i_feat_parameters))


if __name__ == '__main__':
    test_all()



import copy
import datetime
import sys

from geoformat_lib.conf.error_messages import python_inferior_to_3_7_forbidden

from geoformat_lib.conf.fields_variable import (
    geoformat_field_type_to_python_type,
    recast_black_list
)
from geoformat_lib.conf.format_data import (
    value_to_iterable_value,
    is_hexadecimal
)

from geoformat_lib.conversion.metadata_conversion import reorder_metadata_field_index_after_field_drop


def update_field_index(fields_metadata, field_name, new_index):
    """
    This function allow to change field's index position.
    
    :param fields_metadata: geolayer fields metadata
    :param field_name: field_name that we want to re index
    :param new_index: new index
    :return: fields_metadata updated
    """
    output_fields_metadata = copy.deepcopy(fields_metadata)
    # check if field exists in output_fields_metadata
    if field_name in output_fields_metadata:
        # if "index" key is in output_fields_metadata
        if 'index' in output_fields_metadata[field_name]:
            field_name_original_index = output_fields_metadata[field_name]['index']
        # if not we create it
        else:
            for i_field, field_name_in_metadata in enumerate(output_fields_metadata):
                output_fields_metadata[field_name_in_metadata]['index'] = int(i_field)
                if field_name_in_metadata == field_name:
                    field_name_original_index = i_field

        if new_index != field_name_original_index:
            # check if new index is superior or equal than 0
            if new_index >= 0:
                # check if new index is not superior than existing
                if new_index < len(output_fields_metadata):
                    for i_field, (field_name_in_metadata, field_metadata) in enumerate(output_fields_metadata.items()):
                        # get index value for field_name_in_metadata / create it if not exists
                        field_name_in_metadata_idx = field_metadata['index']

                        # update index for field name
                        if field_name_in_metadata == field_name:
                            output_fields_metadata[field_name_in_metadata]['index'] = new_index
                        else:
                            # re index other fields
                            # minus index when field index is between field_name_original_index and new_index
                            if field_name_original_index < field_name_in_metadata_idx <= new_index:
                                output_fields_metadata[field_name_in_metadata]['index'] -= 1
                                field_name_in_metadata_idx -= 1

                            # plus indexes when field index is between new_index and field_name_original_index
                            if new_index <= field_name_in_metadata_idx < field_name_original_index:
                                output_fields_metadata[field_name_in_metadata]['index'] += 1
                else:
                    raise Exception('new index filed name cannot be superior to {nb_fields}'.format(
                        nb_fields=len(output_fields_metadata)-1))
            else:
                raise Exception('new index for field name {field_name} must be superior or equal than 0'.format(
                    field_name=field_name))
    else:
        raise Exception('field name {field_name} not exists'.format(field_name=field_name))

    return output_fields_metadata


def recast_field_value(
    field_value,
    recast_value_to_python_type,
    resize_value_width,
    resize_value_precision
):
    """
    Recast a typed value (field_value) to another typed (recast_to_type) value.

    :param field_value: field value in feature
    :param recast_value_to_python_type: recast field_value to python type
    :param resize_value_width: new width for value
    :param resize_value_precision: new precision for value
    :return: recast value
    """
    # put value in list
    if isinstance(field_value, list):
        field_value_list = field_value
        return_to_list = True
    else:
        field_value_list = [field_value]
        return_to_list = False

    for i_value, field_value in enumerate(field_value_list):
        field_value_type = type(field_value)
        # if value is a list we iterate overt it
        if field_value_type == list:
            field_value_list[i_value] = recast_field_value(
                field_value=field_value,
                recast_value_to_python_type=recast_value_to_python_type
            )
        else:
            # we cannot retype a None value
            if field_value is not None:
                # recast value
                if field_value_type != recast_value_to_python_type:
                    # recast bytes
                    if recast_value_to_python_type is bytes and field_value_type is str:
                        # if field_value is hexadecimal we convert it to bytes
                        if is_hexadecimal(field_value):
                            field_value = bytes.fromhex(field_value)
                        else:
                            field_value = eval(field_value)
                    elif recast_value_to_python_type is str and field_value_type is bytes:
                        # recast bytes to hexadecimal string
                        field_value = field_value.hex()
                    # recast date, time datetime value
                    elif recast_value_to_python_type in {datetime.date, datetime.time, datetime.datetime} and field_value_type is str:
                        if sys.version_info >= (3, 7):
                            field_value = recast_value_to_python_type.fromisoformat(field_value)
                        else:
                            raise Exception(python_inferior_to_3_7_forbidden)
                    elif recast_value_to_python_type is datetime.date and field_value_type is datetime.datetime:
                        field_value = field_value.date()
                    elif recast_value_to_python_type is datetime.time and field_value_type is datetime.datetime:
                        field_value = field_value.time()
                    elif recast_value_to_python_type is datetime.datetime and field_value_type is datetime.date:
                        field_value = datetime.datetime.combine(field_value, datetime.time.min)
                    # recast boolean
                    elif recast_value_to_python_type is bool and field_value_type is str and field_value.upper() == 'FALSE':
                        field_value = False
                    else:
                        try:
                            field_value = recast_value_to_python_type(field_value)
                        except ValueError:
                            raise ValueError('value : {field_value} is not compatible with '
                                             '{recast_value_to_python_type} type'.format(
                                field_value=field_value,
                                recast_value_to_python_type=recast_value_to_python_type)
                            )

                # change precision
                if recast_value_to_python_type == float and resize_value_precision:
                    field_value = round(field_value, resize_value_precision)
                # change width
                if recast_value_to_python_type in {str, float} and resize_value_width:
                    if recast_value_to_python_type is float:
                        field_value = str(field_value)
                        resize_value_width += 1  # add comma width
                    field_value = field_value[:resize_value_width]
                    if recast_value_to_python_type is float:
                        field_value = float(field_value)

            # save value
            field_value_list[i_value] = field_value

    if return_to_list:
        output_value = field_value_list
    else:
        output_value = field_value_list[0]

    return output_value


def recast_field(
    geolayer_to_recast,
    field_name_to_recast,
    recast_to_geoformat_type=None,
    rename_to=None,
    resize_width=None,
    resize_precision=None,
    reindex=None
):
    """
    This function allow to recast a field.

    :param geolayer_to_recast: geolayer that you want to recast
    :param field_name_to_recast: field name that we want to recast
    :param recast_to_geoformat_type: if you want to change field type enter a new field type
    :param rename_to: if you want rename field enter new field name
    :param resize_width: if you want to resize width enter new width
    :param resize_precision: if you want to change field precision enter new precision
    :param reindex: if you want to change field index position enter new index.
    :return: field recasted on geolayer
    """

    # update metadata
    geolayer_to_recast = copy.deepcopy(geolayer_to_recast)
    input_field_metadata = geolayer_to_recast['metadata']['fields'][field_name_to_recast]
    input_field_type = input_field_metadata['type']
    # check if recasting is compatible with actual type
    if recast_to_geoformat_type in recast_black_list[input_field_type]:
        raise Exception("Input type {input_type} cannot recast to {recast_type} type".format(
            input_type=input_field_type,
            recast_type=recast_to_geoformat_type
        ))

    if recast_to_geoformat_type is not None:
        output_field_metadata = {'type': recast_to_geoformat_type}
    else:
        output_field_metadata = {'type': input_field_metadata['type']}

    # width
    if output_field_metadata['type'] in {'Real', 'RealList', 'String', 'StringList'}:
        if resize_width is not None:
            output_field_metadata['width'] = resize_width
        else:
            if "width" in input_field_metadata:
                output_field_metadata['width'] = input_field_metadata['width']
                # if we swap from Real to String  because of the comma width must take +1
                if 'Real' in input_field_metadata['type'] and 'String' in output_field_metadata['type']:
                    output_field_metadata['width'] += 1
                resize_width = output_field_metadata['width']
            else:
                raise Exception("resize_width must be filled for type : {data_type}".format(
                    data_type=output_field_metadata['type']))

    # precision
    if output_field_metadata['type'] in {'Real', 'RealList'}:
        if resize_precision is not None:
            output_field_metadata['precision'] = resize_precision
        else:
            if "precision" in input_field_metadata:
                output_field_metadata['precision'] = input_field_metadata['precision']
                resize_precision = output_field_metadata['precision']
            else:
                raise Exception(
                    "resize_precision must be filled for {data_type} type".format(
                        data_type=output_field_metadata['type']))

    # add index if exists
    if 'index' in input_field_metadata:
        output_field_metadata['index'] = input_field_metadata['index']

    # write output metadata in geolayer_to_recast
    if input_field_metadata != output_field_metadata:
        geolayer_to_recast['metadata']['fields'][field_name_to_recast] = output_field_metadata

    # re index
    if reindex is not None:
        output_fields_metadata = update_field_index(
            fields_metadata=geolayer_to_recast['metadata']['fields'],
            field_name=field_name_to_recast,
            new_index=reindex
        )
        geolayer_to_recast['metadata']['fields'] = output_fields_metadata

    # rename field in metadata
    if rename_to is not None:
        if rename_to in geolayer_to_recast['metadata']['fields']:
            raise Exception('field {rename_to} still exists on geolayer'.format(rename_to=rename_to))
        if rename_to != field_name_to_recast:
            geolayer_to_recast['metadata']['fields'][rename_to] = geolayer_to_recast['metadata']['fields'][
                field_name_to_recast]
            del geolayer_to_recast['metadata']['fields'][field_name_to_recast]

    # define python type to recast
    recast_to_python_type = geoformat_field_type_to_python_type[output_field_metadata['type']]

    # value in list or not
    if output_field_metadata['type'] in {'IntegerList', 'RealList', 'StringList'}:
        in_list = True
    else:
        in_list = False

    # loop over features
    for i_feat, feature in geolayer_to_recast['features'].items():
        # if attributes in feature
        if 'attributes' in feature:
            if field_name_to_recast in feature['attributes']:
                feature_field_name_value = feature['attributes'][field_name_to_recast]
                feature_field_name_type = type(feature_field_name_value)

                if feature_field_name_type == list:
                    # if value in list and output is not in list we have to recast value in str
                    if in_list is False:
                        feature_field_name_value = str(feature_field_name_value)
                else:
                    # if result must be in list
                    if in_list is True:
                        try:
                            eval_feature_field_name_value = eval(feature_field_name_value)
                        except (TypeError, NameError, ValueError):
                            eval_feature_field_name_value = [feature_field_name_value]

                        if isinstance(eval_feature_field_name_value, list):
                            feature_field_name_value = eval_feature_field_name_value
                        else:
                            feature_field_name_value = [feature_field_name_value]

                # recast value(s)
                feature_field_name_value = recast_field_value(
                    field_value=feature_field_name_value,
                    recast_value_to_python_type=recast_to_python_type,
                    resize_value_width=resize_width,
                    resize_value_precision=resize_precision
                )

                # saving recasting data
                feature['attributes'][field_name_to_recast] = feature_field_name_value

                # update field_name change
                if rename_to is not None:
                    feature['attributes'][rename_to] = feature['attributes'][field_name_to_recast]
                    del feature['attributes'][field_name_to_recast]

    return geolayer_to_recast


def drop_field(geolayer, field_name_to_drop):
    """
    This function allow to drop a field in geolayer.

    :param geolayer: input geolayer.
    :param field_name_to_drop: field name in geolayer to drop.
    :return: geolayer with field drop.
    """

    # initialize input
    field_name_to_drop = value_to_iterable_value(value=field_name_to_drop, output_iterable_type=list)
    reorder_field_index = []
    # delete field in metadata
    for field_name in field_name_to_drop:
        if field_name in geolayer['metadata']['fields']:
            if "index" in geolayer['metadata']['fields'][field_name]:
                reorder_field_index.append(geolayer['metadata']['fields'][field_name]['index'])
            # delete field
            del geolayer['metadata']['fields'][field_name]

    # reorder field index
    if reorder_field_index:
        geolayer['metadata']['fields'] = reorder_metadata_field_index_after_field_drop(
            fields_metadata=geolayer['metadata']['fields'],
            reorder_field_index=reorder_field_index
        )

    # delete field in features
    for i_feat, feature in geolayer['features'].items():
        if 'attributes' in feature:
            for field_name in field_name_to_drop:
                if field_name in feature['attributes']:
                    del feature['attributes'][field_name]

    return geolayer



import decimal


def deduce_rounding_value_from_float(float_value):

    d = decimal.Decimal(str(float_value))

    return abs(d.as_tuple().exponent)


def deduce_precision_from_round(rounding_value):

    precision_value = (1 / 10 ** rounding_value)

    return precision_value



import json

from geoformat_lib.conf.fields_variable import geoformat_field_type_to_geojson_field_type
from geoformat_lib.conf.format_data import value_to_iterable_value
from geoformat_lib.conf.geometry_variable import GEOFORMAT_GEOMETRY_TYPE
from geoformat_lib.conversion.feature_conversion import features_filter, feature_list_to_geolayer
from geoformat_lib.conversion.fields_conversion import recast_field_value

from geoformat_lib.conf.path import path_to_file_path, verify_input_path_is_file

from geoformat_lib.conversion.feature_conversion import feature_deserialize


def feature_attributes_to_properties(attributes, recast_field_mapping=None):
    """
    Transforms feature's attributes to geojson properties

    :param attributes: feature attributes
    :param recast_field_mapping: config dict to recast field to geojson properties format
    :return: formatted geojson properties
    """
    if attributes:
        # check is data are serialized or not
        properties = dict(attributes)

        # if recast_field_mapping
        if recast_field_mapping is not None:
            for field_name, field_name_mapping in recast_field_mapping.items():
                field_value_to_recast = attributes.get(field_name, None)
                if field_value_to_recast:
                    field_value_recast = recast_field_value(field_value=field_value_to_recast, **field_name_mapping)
                    properties[field_name] = field_value_recast
    else:
        properties = {}

    return properties


def geoformat_feature_to_geojson_feature(feature, i_feat=None, recast_field_mapping=None):
    """
    Make transformation between geoformat feature to geojson feature.

    :param feature: geoformat feature
    :param i_feat: geoformat feature's id
    :param recast_field_mapping:
    :return: dict like geojson feature
    """
    # get geometries
    geometry = dict(feature.get('geometry', {}))
    # delete bbox if exists
    geometry.pop('bbox', None)
    # get attributes and clean it
    attributes = dict(feature.get('attributes', {}))
    properties = feature_attributes_to_properties(attributes, recast_field_mapping)
    geojson_feature = dict(
        type='Feature',
        geometry=geometry,
        properties=properties
    )
    if i_feat is not None:
        geojson_feature['id'] = i_feat

    return geojson_feature


def geolayer_to_geojson(geolayer, path, overwrite=False, add_extension=False, indent=4):
    """
    Save geolayer to geojson format

    :param geolayer: input geolayer.
    :param path:  output geojson file or directory path .
    :param overwrite: True we overwrite if exists false we raise an exception.
    :param add_extension: True we add .geojson extension to file path if extension .geojson not exists
    :param indent: format output file with specific indentation.
    """

    if add_extension:
        add_extension = '.geojson'

    # check input path and deduce output path
    output_path = path_to_file_path(
        path=path,
        geolayer_name=geolayer['metadata']['name'],
        overwrite=overwrite,
        add_extension=add_extension
    )

    # scan field metadata to check if field must have to be recast
    fields_metadata = geolayer['metadata'].get('fields', {})
    recast_field_mapping = {}
    for field_name, field_metadata in fields_metadata.items():
        recast_to_python_type = geoformat_field_type_to_geojson_field_type.get(field_metadata['type'], None)
        if recast_to_python_type is not None:
            recast_field_mapping[field_name] = {
                "recast_value_to_python_type": recast_to_python_type,
                "resize_value_width": None,
                "resize_value_precision": None
            }

    geojson_features_list = [None] * len(geolayer['features'])
    for i, (i_feat, feature) in enumerate(geolayer['features'].items()):
        if geolayer['metadata'].get('serialize', False):
            feature = feature_deserialize(feature)
        geojson_feature = geoformat_feature_to_geojson_feature(
            feature=feature,
            i_feat=i_feat,
            recast_field_mapping=recast_field_mapping)
        geojson_features_list[i] = geojson_feature

    # create FeatureCollection
    name = geolayer['metadata']['name']
    geojson_feature_collection = {'type': "FeatureCollection", "name": name, "features": geojson_features_list}

    # if reference coordinate system (crs) is filled
    if 'geometry_ref' in geolayer['metadata']:
        if 'crs' in geolayer['metadata']['geometry_ref']:
            geolayer_crs = geolayer['metadata']['geometry_ref']['crs']
            crs_string = {
                "type": "name",
                "properties": {
                    "name": "urn:ogc:def:crs:EPSG::{crs}".format(crs=geolayer_crs)
                }
            }
            crs_string = crs_string
            geojson_feature_collection['crs'] = crs_string

    # write geojson file
    with open(output_path, 'w') as geojson_file:
        json.dump(obj=geojson_feature_collection, fp=geojson_file, indent=indent)


def load_json_data(path):
    """
    Return a tuple that contain : json load to str (index 0) and name of geojson (index 1)

    :param path: json path
    :return: json load to str (index 0) and name of geojson (index 1)
    """
    # verify that path exist
    p = verify_input_path_is_file(path)

    # open file
    with open(path, 'r') as geojson_file:
        json_object = json.loads(geojson_file.read())

    return json_object, p.stem


def json_object_to_feature_generator(
        json_object,
):
    """
    Transform geojson FeatureCollection / Feature or Geometry to geoformat Feature

    :param json_object: geojson object transform to python dict
    :return: geoformat feature
    """
    # if it's a Feature Collection
    if json_object["type"] == "FeatureCollection":
        for i_feat, json_feature in enumerate(json_object["features"]):
            for feature in json_object_to_feature_generator(json_feature):
                if feature:
                    yield feature

    # if it's a Feature
    elif json_object["type"] == "Feature":
        feature = {}

        if "properties" in json_object:
            if json_object["properties"]:
                feature["attributes"] = dict(json_object["properties"])

        if "geometry" in json_object:
            if json_object["geometry"]:
                feature["geometry"] = dict(json_object["geometry"])

        if feature:
            yield feature

    # if it's a geometry
    elif json_object["type"].upper() in GEOFORMAT_GEOMETRY_TYPE:
        feature = {"geometry": dict(json_object)}

        if feature:
            yield feature
    else:
        raise Exception("no geojson compatible data")


# def geojson_to_geocontainer(
#     path,
#     field_name_filter=None,
#     bbox_extent=True,
#     bbox_filter=None,
#     feature_serialize=False,
#     feature_limit=None,
#     feature_offset=None,
# ):
#     """
#
#     """
#
#     ## verify if path is a file or directory path
#     if isinstance(path, (list, tuple, set)):
#         geojson_path_list = path
#     elif os.path.isdir(path):
#         # test relative path
#         dirname = os.path.dirname(__file__)
#         path = os.path.join(dirname, path)
#         # recuperate all geojson file in directory
#         geojson_path_list = glob.glob(path + "*.geojson")
#     elif os.path.isfile(path):
#         geojson_path_list = [path]
#     else:
#         sys.exit("error: your path must be a dir/file path or a list of path")
#
#     # init geocontainer
#     geocontainer = {"layers": {}, "metadata": {}}
#     # is usefull to determine if there is a geometric geolayer in geocontainer (to compute bbox_extent)
#
#     # init GEOMETRIC option
#     geometry_first = True
#     geocontainer_extent = False
#     geometry_in_geocontainer = False
#     geometry_in_geolayer = False
#     for path in geojson_path_list:
#         # recuperate geolayer
#         geolayer = geojson_to_geolayer(
#             path,
#             field_name_filter=field_name_filter,
#             bbox_extent=bbox_extent,
#             bbox_filter=bbox_filter,
#             feature_serialize=feature_serialize,
#         )
#         geolayer_name = geolayer["metadata"]["name"]
#         # GEOMETRY TEST
#         # test if geometry is present in geolayer
#         if "geometry_ref" in geolayer["metadata"]:
#             geometry_in_geolayer = True
#             geometry_in_geocontainer = True
#
#             # if bbox_extent and geometry is present in geolayer
#             if bbox_extent and geometry_first and geometry_in_geolayer:
#                 geocontainer_extent = geolayer["metadata"]["geometry_ref"]["extent"]
#                 geometry_first = False
#             # if bbox_extent and not first
#             if bbox_extent and not geometry_first and geometry_in_geolayer:
#                 geocontainer_extent = bbox_union(
#                     geocontainer_extent, geolayer["metadata"]["geometry_ref"]["extent"]
#                 )
#
#         geocontainer["layers"][geolayer_name] = geolayer
#
#     if bbox_extent and geometry_in_geocontainer:
#         geocontainer["metadata"]["extent"] = geocontainer_extent
#
#     # if geolayer in geocontaire
#     if len(geocontainer["layers"]) > 0:
#         return geocontainer


def from_geojson_get_features_list(
        geojson_in_dict,
        field_name_filter,
        geometry_type_filter,
        bbox_filter,
        serialize,
        bbox_extent,
        feature_limit,
        feature_offset
):
    """
    From geojson transform to dict this function return a features list with some filter option(s)

    :param geojson_in_dict:
    :param field_name_filter:
    :param geometry_type_filter:
    :param bbox_filter:
    :param serialize:
    :param bbox_extent:
    :param feature_limit:
    :param feature_offset:
    :return: list with feature(s) inside
    """
    # create feature generator from file
    features_generator = json_object_to_feature_generator(
        json_object=geojson_in_dict
    )
    # get features list
    features_list = [
        feature for feature in features_filter(
            geolayer_feature_list_or_generator=features_generator,
            field_name_filter=field_name_filter,
            geometry_type_filter=geometry_type_filter,
            bbox_filter=bbox_filter,
            serialize=serialize,
            bbox_extent=bbox_extent,
            feature_limit=feature_limit,
            feature_offset=feature_offset,
        )
    ]

    return features_list


def geojson_to_geolayer(
    path,
    field_name_filter=None,
    geometry_type_filter=None,
    bbox_extent=True,
    bbox_filter=None,
    serialize=False,
    feature_limit=None,
    feature_offset=None,
    force_field_conversion=False,
    crs=None
):
    """
    Convert geojson file to geolayer

    :param path: path to geojson file
    :param field_name_filter: field_name that we want to keep in geolayer (can be a list).
    :param geometry_type_filter: keep only features with geometry type in this variable (can be a list).
    :param bbox_extent: add "bbox" key in each features and "extent" key in geometry metadata.
    :param bbox_filter: keep only feature that intersects bbox (can be a list of bbox).
    :param serialize: True if features in geolayer are serialized (can reduce performance) / False if not.
    :param feature_limit: constrains the number of rows returned in output geolayer.
    :param feature_offset: skip feature before beginning to given line number.
    :param force_field_conversion:  True if you want to force value in field (can change field type) / False if you want
           to deduce field type without forcing field type.
    :param crs: espg code for coordinates reference system.
    :return geolayer: geolayer
    """

    # init and prepare variable
    field_name_filter = value_to_iterable_value(field_name_filter, output_iterable_type=list)
    geometry_type_filter = value_to_iterable_value(geometry_type_filter, output_iterable_type=set)
    bbox_filter = value_to_iterable_value(bbox_filter, output_iterable_type=tuple)

    # load data and get name
    data_geojson_in_dict, name = load_json_data(path)
    if data_geojson_in_dict:
        # filter features
        features_list = from_geojson_get_features_list(
            geojson_in_dict=data_geojson_in_dict,
            field_name_filter=field_name_filter,
            geometry_type_filter=geometry_type_filter,
            bbox_filter=bbox_filter,
            serialize=None,
            bbox_extent=False,
            feature_limit=feature_limit,
            feature_offset=feature_offset
        )

        # create geolayer from filter feature list
        geolayer = feature_list_to_geolayer(
            feature_list=features_list,
            geolayer_name=name,
            field_name_filter=None,
            force_field_conversion=force_field_conversion,
            geometry_type_filter=None,
            bbox_filter=None,
            bbox_extent=bbox_extent,
            crs=crs,
            serialize=serialize,
        )

        # add coordinate reference system
        if "crs" in data_geojson_in_dict and crs is None:
            if 'geometry_ref' in geolayer['metadata']:
                geolayer['metadata']['geometry_ref']['crs'] = int(data_geojson_in_dict['crs']['properties']['name'].split(':')[-1])
        if crs and 'geometry_ref' in geolayer['metadata']:
            geolayer['metadata']['geometry_ref']['crs'] = crs

        return geolayer



from pathlib import Path

from tests.test_all import test_function

from geoformat_lib.conf.path import (
    add_extension_path,
    path_to_file_path,
    verify_input_path_is_file
)

from geoformat_lib.conf.error_messages import path_not_valid

add_extension_path_parameters = {
    0: {
            "path": Path('data/geojson/test'),
            "add_extension": None,
            "return_value": Path('data/geojson/test')
    },
    1: {
            "path": Path('data/geojson/test'),
            "add_extension": '.geojson',
            "return_value": Path('data/geojson/test.geojson')
    },
    2: {
            "path": Path('data/geojson/test.geojson'),
            "add_extension": '.kml',
            "return_value": Path('data/geojson/test.geojson.kml')
    }
}

verify_input_path_is_file_parameters = {
    0: {
        "path": Path(__file__).parent.parent.parent.parent.joinpath('geoformat_lib/conf/path.py'),
        "return_value": Path(__file__).parent.parent.parent.parent.joinpath('geoformat_lib/conf/path.py')
    },
    1: {
        "path": Path(__file__).parent.parent.parent.parent.joinpath('geoformat_lib/conf'),
        "return_value": path_not_valid.format(path=Path(__file__).parent.parent.parent.parent.joinpath('geoformat_lib/conf'))
    },
    2: {
        "path": Path(__file__).parent.parent.parent.parent.joinpath('geoformat_lib/foo'),
        "return_value": path_not_valid.format(path=Path(__file__).parent.parent.parent.parent.joinpath('geoformat_lib/foo'))
    }
}

path_to_file_path_parameters = {
    0: {
        "path": Path(__file__).parent.parent.parent.joinpath('data').as_posix(),
        "geolayer_name": 'test',
        "overwrite": True,
        "add_extension": None,
        "return_value": Path(__file__).parent.parent.parent.joinpath('data/test')
    },
    1: {
        "path": Path(__file__).parent.parent.parent.joinpath('data/test').as_posix(),
        "geolayer_name": 'test',
        "overwrite": True,
        "add_extension": None,
        "return_value": Path(__file__).parent.parent.parent.joinpath('data/test')
    },
    2: {
        "path": Path(__file__).parent.parent.parent.joinpath('data/test').as_posix(),
        "geolayer_name": 'test',
        "overwrite": True,
        "add_extension": '.geojson',
        "return_value": Path(__file__).parent.parent.parent.joinpath('data/test.geojson')
    },
    3: {
        "path": Path(__file__).parent.parent.parent.joinpath('data'),
        "geolayer_name": 'test',
        "overwrite": True,
        "add_extension": None,
        "return_value": Path(__file__).parent.parent.parent.joinpath('data/test')
    },
    4: {
        "path": Path(__file__).parent.parent.parent.joinpath('data/test'),
        "geolayer_name": 'test',
        "overwrite": True,
        "add_extension": None,
        "return_value": Path(__file__).parent.parent.parent.joinpath('data/test')
    },
    5: {
        "path": Path(__file__).parent.parent.parent.joinpath('data/test'),
        "geolayer_name": 'test',
        "overwrite": True,
        "add_extension": '.geojson',
        "return_value": Path(__file__).parent.parent.parent.joinpath('data/test.geojson')
    }
}


def test_all():

    # add_extension_path
    print(test_function(add_extension_path, add_extension_path_parameters))

    # verify_input_path_is_file
    print(test_function(verify_input_path_is_file, verify_input_path_is_file_parameters))

    # path_to_file_path
    print(test_function(path_to_file_path, path_to_file_path_parameters))


if __name__ == '__main__':
    test_all()



from setuptools import setup, find_packages

from geoformat import version

markdown_path = 'README.md'

with open(markdown_path) as read_me_file:
    read_me_file_txt = read_me_file.read()

setup(
    name='geoformat',
    version=str(version(verbose=False)),
    url='https://framagit.org/Guilhain/Geoformat',
    license='MIT',
    author='Guilhain Averlant',
    author_email='g.averlant@mailfence.com',
    description='Geoformat is a GDAL/OGR library overlayer',
    long_description=read_me_file_txt,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=["data", "tests"]),
    # packages=['geoformat'],
    py_modules=[
        'geoformat',
        'requirements',
        # 'geoformat_lib.conf.decorator',
        # 'geoformat_lib.conf.driver_variable',
        # 'geoformat_lib.conf.error_messages',
        # 'geoformat_lib.conf.fields_variable',
        # 'geoformat_lib.conf.format_data',
        # 'geoformat_lib.conf.geometry_variable',
        # 'geoformat_lib.conf.path',
        # 'geoformat_lib.conf.timer',
        # 'geoformat_lib.constraints.primary_key',
        # 'geoformat_lib.conversion.bbox_conversion',
        # 'geoformat_lib.conversion.bytes_conversion',
        # 'geoformat_lib.conversion.coordinates_conversion',
        # 'geoformat_lib.conversion.feature_conversion',
        # 'geoformat_lib.conversion.fields_conversion',
        # 'geoformat_lib.conversion.geolayer_conversion',
        # 'geoformat_lib.conversion.geometry_conversion',
        # 'geoformat_lib.conversion.metadata_conversion',
        # 'geoformat_lib.conversion.precision_tolerance_conversion',
        # 'geoformat_lib.conversion.segment_conversion',
        # 'geoformat_lib.db.db_request',
        # 'geoformat_lib.driver.ogr.ogr_driver',
        # 'geoformat_lib.driver.geojson_driver',
        # 'geoformat_lib.driver.postgresql_driver',
        # 'geoformat_lib.explore_data.print_data',
        # 'geoformat_lib.explore_data.random_geometry',
        # 'geoformat_lib.geoprocessing.connectors.operations',
        # 'geoformat_lib.geoprocessing.connectors.predicates',
        # 'geoformat_lib.geoprocessing.geoparameters.bbox',
        # 'geoformat_lib.geoprocessing.geoparameters.boundaries',
        # 'geoformat_lib.geoprocessing.geoparameters.lines',
        # 'geoformat_lib.geoprocessing.matrix.adjacency',
        # 'geoformat_lib.geoprocessing.measure.area',
        # 'geoformat_lib.geoprocessing.measure.distance',
        # 'geoformat_lib.geoprocessing.measure.length',
        # 'geoformat_lib.geoprocessing.area',
        # 'geoformat_lib.geoprocessing.length',
        # 'geoformat_lib.geoprocessing.line_merge',
        # 'geoformat_lib.geoprocessing.merge_geometries',
        # 'geoformat_lib.geoprocessing.point_on_linestring',
        # 'geoformat_lib.geoprocessing.split',
        # 'geoformat_lib.geoprocessing.union',
        # 'geoformat_lib.index.attributes.hash',
        # 'geoformat_lib.index.geometry.grid',
        # 'geoformat_lib.obj.geocontainer',
        # 'geoformat_lib.obj.geometry',
        # 'geoformat_lib.processing.data.clauses',
        # 'geoformat_lib.processing.data.field_statistics'
    ],
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: GIS",
    ]
)



from geoformat_lib.geoprocessing.connectors.operations import (
    coordinates_to_point
)

def format_coordinates(
    coordinates_list_tuple,
    format_to_type=list,
    precision=None,
    delete_duplicate_following_coordinates=False
):
    """
    This function allow to :
        - change type of coordinate's storage
        - round coordinates
        - delete duplicate following coordinates in


    :param coordinates_list_tuple: tuple or list containing coordinates
    :param format_to_type: coordinate storage data type (list (by default) or tuple)
    :param precision: desired number of decimal for coordinates
    :param delete_duplicate_following_coordinates: option to delete similar coordinates that follow each other
    :return: formatted coordinates
    """
    # TODO add delete duplicated part
    if coordinates_list_tuple:
        output_coordinates_list_tuple = [None] * len(coordinates_list_tuple)
        for i_coord, coordinates in enumerate(coordinates_list_tuple):
            if isinstance(coordinates, (int, float)):
                if precision is not None:
                    coordinates = round(coordinates, precision)
                coord_tuple = coordinates
            elif isinstance(coordinates, (list, tuple)):
                # here we first format type of storage of coordinates and precision
                coord_tuple = format_coordinates(
                    coordinates_list_tuple=coordinates,
                    format_to_type=format_to_type,
                    precision=precision,
                    delete_duplicate_following_coordinates=delete_duplicate_following_coordinates
                )
            else:
                raise TypeError('must be list or tuple containing int or float')

            # save result
            output_coordinates_list_tuple[i_coord] = coord_tuple

        # DELETE DUPLICATE FOLLOWING COORDINATES
        # in step before if we reformat precision we can create duplicate coordinates then if
        # delete_duplicate_following_coordinates is True we have to make a second scan of coordinates
        # if we are at level of part of geometry (list that contains coordinates in it)
        if delete_duplicate_following_coordinates is True and isinstance(output_coordinates_list_tuple[0], (list, tuple)) and isinstance(
                output_coordinates_list_tuple[0][0], (int, float)):
            # scan list of coordinates or list of list of coordinates (don't scan coordinate list)
            # if coordinates
            retype_to_tuple = False
            # transform tuple to list (because we need to edit coordinates)
            if isinstance(output_coordinates_list_tuple, tuple):
                retype_to_tuple = True
                output_coordinates_list_tuple = list(output_coordinates_list_tuple)

            # try to find duplicate coordinates index
            duplicate_coordinates_idx_list = []
            for idx_inside_coord, inside_coordinates in enumerate(output_coordinates_list_tuple):
                if idx_inside_coord == 0:
                    duplicate_coordinates_idx_list = []
                else:
                    # if coordinates are same that previous index of coordinates is save in
                    # duplicate_coordinates_idx_list (we delete this coordinates after the loop)
                    if previous_coordinates == inside_coordinates:
                        duplicate_coordinates_idx_list.append(idx_inside_coord)
                previous_coordinates = inside_coordinates

            # delete duplicate
            if duplicate_coordinates_idx_list:
                for idx in reversed(duplicate_coordinates_idx_list):
                    del output_coordinates_list_tuple[idx]

            # retype to tuple if output_coordinates_list_tuple is originaly a tuple
            if retype_to_tuple is True:
                output_coordinates_list_tuple = tuple(output_coordinates_list_tuple)

        # FORMAT coordinates storage type
        # reformat coordinates storage type (list or tuple)
        if format_to_type:
            output_coordinates_list_tuple = format_to_type(output_coordinates_list_tuple)
    else:
        if format_to_type == list:
            output_coordinates_list_tuple = []
        else:
            output_coordinates_list_tuple = ()

    return output_coordinates_list_tuple


def coordinates_to_2d_coordinates(coordinates_list):
    """
    Convert a coordinates list with x dimension to 2d dimension list

    :param coordinates_list: list of coordinates with 2 or more dimension
    :return: coordinates list with only 2 dimensions.
    """

    if coordinates_list:
        if isinstance(coordinates_list[0], (int, float)):
            new_coordinates = [coordinates_list[0], coordinates_list[1]]
        elif isinstance(coordinates_list[0], (tuple, list)):
            new_coordinates = [None] * len(coordinates_list)
            for i_coord, under_coordinates in enumerate(coordinates_list):
                new_coordinates[i_coord] = coordinates_to_2d_coordinates(under_coordinates)
        else:
            raise Exception('error your geometry in input is not correct')
    else:
        new_coordinates = []

    return new_coordinates


def coordinates_to_centroid(coordinates_list_tuple, precision=None):
    """
    Return the centroid of given coordinates list or tuple

    :param coordinates_list_tuple: (list or tuple) coordinates list or tuple
    :param precision: (int) precision of coordinates (number of decimal places)
    :return: (tuple) centroid
    """

    for i_point, point in enumerate(coordinates_to_point(coordinates_list_tuple)):
        if i_point == 0:
            mean_point = list(point)
        else:
            mean_point[0] += point[0]
            mean_point[1] += point[1]

    nb_coordinates = i_point + 1
    centroid = [mean_point[0] / nb_coordinates, mean_point[1] / nb_coordinates]

    if precision:
        centroid = format_coordinates(coordinates_list_tuple=centroid, precision=precision)

    return centroid




# python version
python_inferior_to_3_7_forbidden = "you must have a version of python higher or equal to 3.7 to execute this function"

# driver
import_ogr_error = "cannot run this function ! Python-gdal library not exists"
import_psycopg2_error = "cannot run this function ! psycopg2 library not installed"

# path
path_not_valid = "path : {path} is not valid"
path_not_valid_file_exists_overwrite_is_false = "path : {path} exists."

# fields
missing_field = "field : {field_name} does not exists in geolayer"

# values

# non unique value
non_unique_values = "field : {field_name} contains non-unique values"
